from langgraph.graph import StateGraph, END

from .agents import *
from .agent_protocol import MultiAgentState


class Orchestration:
    def __init__(self):
        self.orch = StateGraph(MultiAgentState)

        # Nodes ------------------------
        # Tools selector and requests node
        self.orch.add_node("tools_selector_node", tools_selector_node)
        # Quality assurance node
        self.orch.add_node("judge_node", judge_node)

        # Edges ------------------------
        self.orch.set_entry_point("tools_selector_node")
        self.orch.add_edge("tools_selector_node", "judge_node")
        self.orch.add_conditional_edges(
            "judge_node",
            judge_decision,
            {
                'GOOD': END,
                'BAD': END  # 'tools_selector_node'
            }
        )

        self.graph = self.orch.compile()

    async def invoke(self, prompt: str):
        events = self.graph.astream(
            {"question": prompt},
            {"recursion_limit": 30},
        )

        out = []
        async for s in events:
            out.append(s)
        yield out



import os
from openai import OpenAI
import logging
from asyncio import gather
from pydantic import BaseModel
from typing import Any

from .tools import openai_url_invoke
from .agent_protocol import MultiAgentState
from mcp_master.config import global_config as gconfig
from mcp_master.config import ConfigError

# --------------------------------------------------------------------------------------------
# Config -------------------------------------------------------------------------------------
# --------------------------------------------------------------------------------------------


# Set up logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')


class Config(BaseModel):
    model_id: str = '',
    tools: list = [],
    master_server_client: Any = None,
    dispatcher_system_message: str = '',


config = Config()

# OpenAI chat client
openai_client = None


# --------------------------------------------------------------------------------------------
# Deployment Nodes ---------------------------------------------------------------------------
# --------------------------------------------------------------------------------------------


async def tools_selector_node(state: MultiAgentState):
    global openai_client

    logging.info(f'Selecting tools from {[tool['function']['name'] for tool in config.tools]}...')

    # Exit for invalid selector_model_id values
    config.model_id = gconfig.selector_model_id
    if config.model_id is None or len(config.model_id) == 0:
        raise ConfigError(f'Ensure your selector_model_id is properly configured via set_config(). It is currently {config.model_id}.')

    # Exit for invalid API key
    if gconfig.OPENAI_API_KEY is None or len(gconfig.OPENAI_API_KEY) == 0:
        raise ConfigError(f'Ensure your OPENAI_API_KEY is properly configured via set_config(). It is currently {gconfig.OPENAI_API_KEY}.')
    os.environ["OPENAI_API_KEY"] = gconfig.OPENAI_API_KEY

    # Exit for empty API key
    if gconfig.OPENAI_BASE_URL is not None:
        if len(gconfig.OPENAI_BASE_URL) == 0:
            raise ConfigError(
                f'Ensure your OPENAI_BASE_URL is properly configured via set_config(). It is currently empty.')

        # Save API key to environment variables
        os.environ["OPENAI_BASE_URL"] = gconfig.OPENAI_BASE_URL

    # Initialize openAI chat client
    if openai_client is None:
        openai_client = OpenAI()

    # Select tools
    messages = [
        {"role": "system", "content": config.dispatcher_system_message},
        {"role": "user", "content": state.question},
    ]

    response = openai_client.chat.completions.create(
        model=config.model_id,
        max_tokens=3000,
        messages=messages,
        tools=config.tools
    )

    tool_calls = response.choices[0].message.tool_calls
    logging.info(f'Selected tools: {tool_calls}')

    # Call selected tools
    if tool_calls:
        tool_callables = [
            config.master_server_client.call_tool(
                tool.function.name, eval(tool.function.arguments)
            )
            for tool in tool_calls
        ]

        results = await gather(*tool_callables)
        external_data = [result.content[0].text for result in results]

        logging.info(f'Tool results: {external_data}')
        return {'tools_requested': tool_calls, 'external_data': str(external_data)}

    # Safety in case the model chooses to generate its own response
    response = response.choices[0].message.content
    logging.info(f'Model-generated response: {response}')
    return {'tools_requested': None, 'external_data': response}


def judge_node(state: MultiAgentState):
    logging.info(f'Judging response...')
    # Answers:
    # GOOD - send to client
    # BAD - try again

    decision, _ = openai_url_invoke(
        gconfig.judge_model_id,
        state.question,
        str(state.external_data),
        f'You are a seasoned expert. Your role is to determine the quality of the answer generated by '
        f'a team of other AI models. Output "GOOD" if the response answers the question and matches '
        f'real world consensus, otherwise output "BAD". Do not output anything else besides those two '
        f'options. Output in all capital letters.',
        gconfig.judge_model_service_url,
    )
    decision = decision.strip().upper()

    logging.info(f'Judge decision: {decision}')
    return {'qa_assessment': decision}


def judge_decision(state: MultiAgentState):
    logging.info(f'Routing judge decision...')

    return state.qa_assessment



from typing import Optional
from mcp import ClientSession
from mcp.client.streamable_http import streamablehttp_client
import os
import asyncio
import httpx
import logging
from subprocess import Popen
from contextlib import AsyncExitStack

from mcp_master.config import global_config as gconfig
from mcp_master.orchestration.agents import config as agent_config


# --------------------------------------------------------------------------------------------
# Config -------------------------------------------------------------------------------------
# --------------------------------------------------------------------------------------------


TOOL_NAME_ORIGIN_SEPARATOR = '09090'
MASTER_TOOL_DESCRIPTION_HEADER = 'Automatically interface with other MCP servers for the following tools:'
MASTER_DISPATCHER_SYSTEM_HEADER = (
    "You are a tool dispatcher agent who decides which tools to dispatch to based on the user's input. Depending on "
    "your answer, question will be routed to the right tools, so your task is crucial."
)

# Set up logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')


# --------------------------------------------------------------------------------------------
# Master Server Client Class -----------------------------------------------------------------
# --------------------------------------------------------------------------------------------


class MasterServerClient:
    def __init__(self, app):
        # All streamable-http sessions by server_filename
        self.sessions = {}

        # All streamable-http streams by server_filename
        self._streams_contexts = {}

        # All streamable-http contexts by server_filename
        self._session_contexts = {}

        # Available tools from each server
        self.available_tools = {}

        # self.available_tools,  disregarding what server each tool is served from
        self.available_tools_flattened = []

        # Popens used to start local sub servers automatically
        self._sub_server_popens = {}

        # FastMCP app
        self._app = app

        #
        self._exit_stack = AsyncExitStack()

    async def check_if_server_running(self, server_url: str, server_filename: str):
        try:
            # Send request to see if the server is already running
            async with httpx.AsyncClient(timeout=30.0) as client:
                logging.info(f"HTTP GET attempt to {server_url}")

                response = await client.get(server_url)
                response.raise_for_status()

                # Exit if already running
                return True

        except httpx.HTTPStatusError:
            # Exit if sent a redirect error (standard behavior)
            return True

        except (httpx.TimeoutException, httpx.ConnectError) as e:
            logging.warning(f"HTTP request attempt failed: {type(e).__name__}: {e}")

            # Exit if the server_path is not absolute
            server_dir = os.path.join(gconfig.autostart_abspath)
            os.chdir(server_dir)
            if not os.path.isabs(server_dir):
                logging.info(f"{server_dir} is not an absolute path.")
                return False

            # Exit if the server is not present in the folder
            server_path = os.path.normpath(os.path.join(server_dir, f'{server_filename}.py'))
            if not os.path.exists(server_path):
                logging.info(f"Unable to find server {server_filename} in {server_path}.")
                return False

            # Run start command to start the server
            logging.info(f"Sending start command for {server_filename}...")
            start_popen = Popen(['python', f'{server_filename}.py'])
            self._sub_server_popens[server_filename] = start_popen

            # Wait for the server to start, abort after 10 seconds
            attempt_count = 1
            async with httpx.AsyncClient(timeout=10.0) as client:
                while True:
                    try:
                        logging.debug(f"HTTP GET attempt to {server_url}")

                        response = await client.get(server_url)
                        response.raise_for_status()

                        # Exit if running
                        return True

                    except (httpx.TimeoutException, httpx.ConnectError) as e:
                        # Try again in 0.5 seconds if the request failed
                        logging.warning(f"HTTP request attempt {attempt_count} failed: {type(e).__name__}: {e}")
                        logging.info(f"Waiting 0.5 seconds...")
                        attempt_count += 1
                        await asyncio.sleep(0.5)

                    except httpx.HTTPStatusError:
                        # Exit if sent a redirect error (normal behavior)
                        return True

                    except Exception as e:
                        # Abort if an unusual error is caught
                        logging.error(f"Unexpected error in HTTP request: {e}")
                        return False

        except Exception as e:
            # Abort if an unusual error is caught
            logging.error(f"Unexpected error in HTTP request: {e}")
            return False

    async def connect_to_server(self, server_url: str, server_filename: str, headers: dict = {}):
        """Connect to an MCP server running on streamable HTTP"""
        # Exit if the server_filename is a duplicate
        if server_filename in self.sessions:
            logging.warning(f"Connection to {server_filename} aborted as it mirrors an existing server_filename.")
            return

        # Exit if the server is not running despite attempts to start it
        if not await self.check_if_server_running(server_url, server_filename):
            logging.error(f"Failed to connect to server {server_filename} at {server_url}.")
            return

        # Initialize client and session
        client = await self._exit_stack.enter_async_context(streamablehttp_client(url=server_url, headers=headers))
        self._streams_contexts[server_filename] = client
        read_stream, write_stream, _ = client

        session = await self._exit_stack.enter_async_context(ClientSession(read_stream, write_stream))
        self._session_contexts[server_filename] = session
        self.sessions[server_filename] = session

        await self.sessions[server_filename].initialize()

        # Save the sub server's available tools
        await self.get_available_tools(server_filename)

        logging.info(f"Connected to server {server_filename} at {server_url}.")

    async def get_available_tools(self, server_filename: str):
        """Get available tools from the server"""
        try:
            # Fetch tools
            logging.info(f"Fetching available server tools from {server_filename}...")
            response = await self.sessions[server_filename].list_tools()
            logging.info(f"Connected to MCP server {server_filename} with tools {[tool.name for tool in response.tools]}.")

            # Format tools for OpenAI
            available_tools = [
                {
                    "type": 'function',
                    "function": {
                        "name": f"{tool.name}{TOOL_NAME_ORIGIN_SEPARATOR}{server_filename}",
                        "description": tool.description,
                        "parameters": tool.inputSchema,
                    },
                    "strict": True,
                }
                for tool in response.tools
            ]

            # Save tools
            self.available_tools[server_filename] = available_tools
            self.available_tools_flattened.extend(available_tools)

            # Compile tool descriptions into the master server tool description
            self.compile_tool_descriptions()

        except Exception as e:
            logging.error(f'Tool fetch for server {server_filename} failed: {e}')

            # Blank list failsafe in case the tool fetch fails
            self.available_tools[server_filename] = []

    def compile_tool_descriptions(self):
        logging.info("Compiling all available tool descriptions...")

        # Generate description based on sub mcp server tool descriptions
        tool_description = MASTER_TOOL_DESCRIPTION_HEADER
        dispatcher_system_message = f'{MASTER_DISPATCHER_SYSTEM_HEADER}\nThere are {len(self.available_tools_flattened)} possible tools to use:'

        for server in self.available_tools:
            for tool in self.available_tools[server]:
                tool_description += f'\n{tool['function']['description']}'
                dispatcher_system_message += f'\n - {tool['function']['name']}: {tool['function']['description']}'

        # Save description directly to tool
        logging.info(f"Compiled description: {tool_description}")
        self._app._tool_manager._tools.get('access_sub_mcp').description = tool_description

        # Save dispatcher node system message to orchestration
        dispatcher_system_message += "Always call at least one tool. Do not attempt to generate your own response to the user's query."
        agent_config.dispatcher_system_message = dispatcher_system_message

    async def call_tool(self, tool_name: str, tool_args: Optional[dict]):
        tool_name, server_filename = tool_name.split(TOOL_NAME_ORIGIN_SEPARATOR)
        logging.info(f"Calling tool {tool_name} from {server_filename} with args {tool_args}...")

        try:
            # Call tool
            result = await self.sessions[server_filename].call_tool(tool_name, tool_args)
            return result
        except Exception:
            # Remove tool from server tool list
            if tool_name in self.available_tools[server_filename]:
                self.available_tools[server_filename].remove(tool_name)

            logging.error(f"Tool {tool_name} from {server_filename} is currently unavailable:")
            raise Exception(e)

    async def server_loop(self):
        while True:
            try:
                await asyncio.sleep(1)
            except Exception as e:
                logging.error(f"\nServer loop error: {str(e)}")

    async def cleanup(self):
        """Properly clean up the sessions and streams"""
        if self._sub_server_popens:
            for popen_id in self._sub_server_popens:
                self._sub_server_popens[popen_id].terminate()

        await self._exit_stack.aclose()

        # if self._session_contexts:
        #     for context_id in self._session_contexts:
        #         await self._session_contexts[context_id].__aexit__(None, None, None)
        #
        # if self._streams_contexts:
        #     for context_id in self._session_contexts:
        #         await self._streams_contexts[context_id].__aexit__(None, None, None)



import logging
from fastmcp import FastMCP
from mcp.server import Server
from starlette.requests import Request
from starlette.applications import Starlette
from mcp.server.sse import SseServerTransport
from starlette.routing import Mount, Route


# Set up logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

# Initialize FastMCP server
app = FastMCP("test2")
started: bool = False

@app.tool(description="Return num1 (even) to the power of num2")
async def even_exponent(num1: int, num2: int) -> str:
    logging.info(f"Finding {num1}^{num2}")
    """
    Return an even number num1 to the power of num2

    Args:
        num1: the base to raise by num2
        num1: the exponent to raise num1 to

    Returns:
        List of dictionaries containing status information
    """

    return num1 ** num2


def create_starlette_app(mcp_server: Server, *, debug: bool = False) -> Starlette:
    """Create a Starlette application that can serve the provided mcp server with SSE."""
    sse = SseServerTransport("/messages/")

    async def handle_sse(request: Request) -> None:
        async with sse.connect_sse(
                request.scope,
                request.receive,
                request._send,  # noqa: SLF001
        ) as (read_stream, write_stream):
            await mcp_server.run(
                read_stream,
                write_stream,
                mcp_server.create_initialization_options(),
            )

    return Starlette(
        debug=debug,
        routes=[
            Route("/sse", endpoint=handle_sse),
            Mount("/messages/", app=sse.handle_post_message),
        ],
    )

def startup():
    global started
    
    if started:
        return

    started = True
    try:
        app.run(transport="streamable-http", host="0.0.0.0", port=8092)
    except KeyboardInterrupt:
        pass
    finally:
        logging.info('Server test_server_2 successfully shut down.')

if __name__ == '__main__':
    startup()



import asyncio
import logging
from fastmcp import FastMCP
from mcp.server import Server
from starlette.requests import Request
from starlette.applications import Starlette
from mcp.server.sse import SseServerTransport
from starlette.routing import Mount, Route

from .master_server_client import MasterServerClient
from mcp_master.orchestration import Orchestration
from mcp_master.orchestration.agents import config as agent_config

# Set up logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')


# Initialize FastMCP server
class MasterMCPServer:
    def __init__(self, port: int = 3000, sub_servers: list[tuple] = []):
        # FastMCP server
        self.app = FastMCP("master_server")

        # Client to access sub-MCP servers
        self.master_server_client = None

        # Hosting port for the master server
        self.port = port

        # ("url", "servername") pairs to connect to servers with
        self.sub_servers = sub_servers

        # Initialize orchestration graph
        self.orch = Orchestration()

        @self.app.tool()
        async def access_sub_mcp(query: str):
            logging.info(f'Collecting tool information for query: {query}')

            agent_config.tools = self.master_server_client.available_tools_flattened
            agent_config.master_server_client = self.master_server_client

            result = await self.orch.graph.ainvoke(
                {"question": query},
                {"recursion_limit": 30},
            )
            logging.info(f'Orchestration result: {result}')

            return result.get('external_data')

    def create_starlette_app(self, mcp_server: Server, *, debug: bool = False) -> Starlette:
        """Create a Starlette application that can serve the provided mcp server with SSE."""
        sse = SseServerTransport("/messages/")

        async def handle_sse(request: Request) -> None:
            async with sse.connect_sse(
                    request.scope,
                    request.receive,
                    request._send,
            ) as (read_stream, write_stream):
                await mcp_server.run(
                    read_stream,
                    write_stream,
                    mcp_server.create_initialization_options(),
                )

        return Starlette(
            debug=debug,
            routes=[
                Route("/sse", endpoint=handle_sse),
                Mount("/messages/", app=sse.handle_post_message),
            ],
        )

    # Server-server communications
    async def initialize_interserver_comms(self):
        self.master_server_client = MasterServerClient(self.app)

        try:
            for server in self.sub_servers:
                await self.master_server_client.connect_to_server(*server)

            await self.master_server_client.server_loop()
        except KeyboardInterrupt:
            pass
        finally:
            await self.master_server_client.cleanup()
            pass

    async def run_app(self):
        try:
            await self.app.run_async(transport="streamable-http", host="0.0.0.0", port=self.port)
        except KeyboardInterrupt:
            pass

    async def _startup(self):
        await asyncio.gather(self.initialize_interserver_comms(), self.run_app())

    def startup(self):
        try:
            asyncio.run(self._startup())
        except KeyboardInterrupt:
            pass
        finally:
            logging.info('Master MCP server successfully shut down.')


if __name__ == "__main__":
    server = MasterMCPServer(
        port=3000,
        sub_servers=[
            ("http://localhost:8091/mcp", 'test_server_1'),
            ("http://localhost:8092/mcp", 'test_server_2')
        ]
    )
    asyncio.run(server.startup())



from mcp_master import MasterMCPServer
from mcp_master import global_config as gconfig
from os import getenv

gconfig.selector_model_id = ''  # Set this to your tool selector model ID
gconfig.judge_model_id = ''  # Set this to your judge model ID
gconfig.judge_model_service_url = ''  # Set this to where your judge LLM is hosted
gconfig.OPENAI_API_KEY = getenv('OPENAI_API_KEY')
gconfig.OPENAI_BASE_URL = getenv('OPENAI_BASE_URL')  # Set this to where your other LLMs will be hosted, None for default

# Create an MCP server on port 3000 with two test servers
# Ensure both test servers are running by starting them in the terminal before starting demo_master_server.py
server = MasterMCPServer(
    port=3000,
    sub_servers=[
        # (server url, server identifier) pairs - ensure all server identifiers are unique
        # If the server is located locally (as the demo servers are), ensure the server identifier matches the server's file name (without the .py)
        ("http://localhost:8091/mcp", 'test_server_1'),
        ("http://localhost:8092/mcp", 'test_server_2')
    ]
)
server.startup()



import requests
from bs4 import BeautifulSoup
from urllib.parse import quote

def generate_medrxiv_search_url(term=None, title=None, author1=None, author2=None, abstract_title=None, 
                                text_abstract_title=None, journal_code="medrxiv", section=None,
                                start_date=None, end_date=None, num_results=10, sort="relevance-rank"):

    base_url = "https://www.medrxiv.org/search/"
    query_parts = []
    if term:
        query_parts.append(f"{quote(term)}")
    if title:
        query_parts.append(f"title%3A{quote(title)} title_flags%3Amatch-all")
    if author1:
        query_parts.append(f"author1%3A{quote(author1)}")
    if author2:
        query_parts.append(f"author2%3A{quote(author2)}")
    if abstract_title:
        query_parts.append(f"abstract_title%3A{quote(abstract_title)} abstract_title_flags%3Amatch-all")
    if text_abstract_title:
        query_parts.append(f"text_abstract_title%3A{quote(text_abstract_title)} text_abstract_title_flags%3Amatch-all")
    if journal_code:
        query_parts.append(f"jcode%3A{quote(journal_code)}")
    if section:
        query_parts.append(f"toc_section%3A{quote(section)}")
    if start_date and end_date:
        query_parts.append(f"limit_from%3A{start_date} limit_to%3A{end_date}")
    
    query_parts.append(f"numresults%3A{num_results}")
    query_parts.append(f"sort%3A{quote(sort)} format_result%3Astandard")

    return base_url + "%20".join(query_parts)

def scrape_medrxiv_results(search_url):
    headers = {
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36'
    }
    response = requests.get(search_url, headers=headers)

    if response.status_code == 200:
        soup = BeautifulSoup(response.text, 'html.parser')
        articles = soup.find_all('li', class_='search-result')

        results = []
        for article in articles:
            title_tag = article.find('span', class_='highwire-cite-title')
            title = title_tag.text.strip() if title_tag else "No title"

            authors_tag = article.find('span', class_='highwire-citation-authors')
            authors = authors_tag.text.strip() if authors_tag else "No authors"

            abstract_tag = article.find('div', class_='highwire-cite-snippet')
            abstract = abstract_tag.text.strip() if abstract_tag else "No abstract"

            link_tag = article.find('a', class_='highwire-cite-linked-title')
            link = "https://www.medrxiv.org" + link_tag['href'] if link_tag else "No link"

            doi_tag = article.find('span', class_='highwire-cite-metadata-doi')
            doi_link = doi_tag.text.strip().replace("doi:", "").strip() if doi_tag else "No DOI"

            metadata = {}
            result = {
                "Title": title,
                "Authors": authors,
                "DOI_link": doi_link,
                "Link": link
            }
            if doi_link != "No DOI":
                metadata = doi_get_medrxiv_metadata(doi_link.replace("https://doi.org/", ""))
                if metadata:
                    result.update(metadata)

            results.append(result)
        
        return results
    else:
        print(f"Error: Unable to fetch data (status code: {response.status_code})")
        return None

def doi_get_medrxiv_metadata(doi, server="medrxiv"):
    """使用 medRxiv API 通过 DOI 获取文章的详细元数据"""
    url = f"https://api.medrxiv.org/details/{server}/{doi}/na/json"
    headers = {
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36'
    }
    
    response = requests.get(url, headers=headers)
    
    if response.status_code == 200:
        data = response.json()
        if 'collection' in data and len(data['collection']) > 0:
            article = data['collection'][0]
            return {
                "DOI": article.get("doi", "No DOI"),
                "Title": article.get("title", "No title"),
                "Authors": article.get("authors", "No authors"),
                "Corresponding Author": article.get("author_corresponding", "No corresponding author"),
                "Corresponding Institution": article.get("author_corresponding_institution", "No institution"),
                "Date": article.get("date", "No date"),
                "Version": article.get("version", "No version"),
                "Category": article.get("category", "No category"),
                "JATS XML Path": article.get("jats xml path", "No XML path"),
                "Abstract": article.get("abstract", "No abstract")
            }
        else:
            print("No data found for DOI:", doi)
            return None
    else:
        print(f"Error: Unable to fetch metadata (status code: {response.status_code})")
        return None

def search_key_words(key_words, num_results=10):
    # 生成搜索 URL
    search_url = generate_medrxiv_search_url(term=key_words, num_results=num_results)

    print("Generated URL:", search_url)

    # 获取并解析搜索结果
    articles = scrape_medrxiv_results(search_url)

    return articles


def search_advanced(term, title, author1, author2, abstract_title, text_abstract_title, section, start_date, end_date, num_results):
    # 生成搜索 URL
    search_url = generate_medrxiv_search_url(term, title=title, author1=author1, author2=author2, 
                                            abstract_title=abstract_title, 
                                            text_abstract_title=text_abstract_title,
                                            section=section, start_date=start_date, 
                                            end_date=end_date, num_results=num_results)

    print("Generated URL:", search_url)

    # 获取并解析搜索结果
    articles = scrape_medrxiv_results(search_url)

    return articles



if __name__ == "__main__":
    # 1. search_key_words 
    key_words = "COVID-19"
    articles = search_key_words(key_words, num_results=5)
    print(articles)

    # 2. search_advanced
    # 示例：用户输入搜索参数
    term = "COVID-19"
    title = "COVID-19"
    author1 = "MacLachlan"
    author2 = None
    abstract_title = None 
    text_abstract_title = None
    section = None 
    start_date = None 
    end_date = None 
    num_results = 5
    articles = search_advanced(term, title, author1, author2, abstract_title, text_abstract_title, section, start_date, end_date, num_results)
    print(articles)

    # 3. doi get medrxiv metadata
    doi = "10.1101/2025.03.09.25323517"
    metadata = doi_get_medrxiv_metadata(doi)
    print(metadata)



import os
from pydantic import BaseModel
from typing import Optional


class GlobalConfig(BaseModel):
    # Tool selection model ID
    selector_model_id: str = ''

    # Selection judge model ID
    judge_model_id: str = ''

    # Judge service URL
    judge_model_service_url: str = ''

    # OpenAI API Key
    OPENAI_API_KEY: str = ''

    # OpenAI Base URL
    OPENAI_BASE_URL: Optional[str] = None

    # Autostart path for locally available servers, defaulted to ../../../examples/demo-servers
    autostart_abspath: str = os.path.normpath(
        os.path.join(os.path.dirname(__file__), '..', '..', 'examples', 'demo-servers')
    )

    # Whether a token is required for access
    private: bool = False


global_config = GlobalConfig()


class ConfigError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)



from typing import Any
from pydantic import BaseModel


class MultiAgentState(BaseModel):
    question: str
    external_data: list = []
    external_summaries: list = []
    data_sources: list = []
    qa_instructions: str = ''
    qa_assessment: str = ''
    answer: str = ''
    tools_requested: list = []

    def __init__(self, /, **data: Any):
        super().__init__(**data)



from typing import Any, List, Dict, Optional
import asyncio
import logging
from fastmcp import FastMCP
from mcp.server import Server
from starlette.requests import Request
from mcp.server.fastmcp.prompts import base
from starlette.applications import Starlette
from mcp.server.sse import SseServerTransport
from starlette.routing import Mount, Route

from medrxiv_web_search import search_key_words, search_advanced, doi_get_medrxiv_metadata

# Set up logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

# Initialize FastMCP server
app = FastMCP("medrxiv")

@app.tool(description="Search medrxiv with keywords")
async def search_medrxiv_key_words(key_words: str, num_results: int = 10) -> List[Dict[str, Any]]:
    logging.info(f"Searching for articles with key words: {key_words}, num_results: {num_results}")
    """
    Search for articles on medRxiv using key words.

    Args:
        key_words: Search query string
        num_results: Number of results to return (default: 10)

    Returns:
        List of dictionaries containing article information
    """

    try:
        results = await asyncio.to_thread(search_key_words, key_words, num_results)
        return results
    except Exception as e:
        return [{"error": f"An error occurred while searching: {str(e)}"}]

@app.tool(description="Search medrxiv with advanced methods")
async def search_medrxiv_advanced(
    term: Optional[str] = None,
    title: Optional[str] = None,
    author1: Optional[str] = None,
    author2: Optional[str] = None,
    abstract_title: Optional[str] = None,
    text_abstract_title: Optional[str] = None,
    section: Optional[str] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    num_results: int = 5
) -> List[Dict[str, Any]]:
    logging.info(f"Performing advanced search with parameters: {locals()}")
    """
    Perform an advanced search for articles on medRxiv.

    Args:
        term: General search term
        title: Search in title
        author1: First author
        author2: Second author
        abstract_title: Search in abstract and title
        text_abstract_title: Search in full text, abstract, and title
        section: Section of medRxiv
        start_date: Start date for search range (format: YYYY-MM-DD)
        end_date: End date for search range (format: YYYY-MM-DD)
        num_results: Number of results to return (default: 10)

    Returns:
        List of dictionaries containing article information
    """
    try:
        results = await asyncio.to_thread(
            search_advanced,
            term, title, author1, author2, abstract_title, text_abstract_title,
            section, start_date, end_date, num_results
        )
        return results
    except Exception as e:
        return [{"error": f"An error occurred while performing advanced search: {str(e)}"}]

@app.tool(description="Search medrxiv with metadata")
async def get_medrxiv_metadata(doi: str) -> Dict[str, Any]:
    logging.info(f"Fetching metadata for DOI: {doi}")
    """
    Fetch metadata for a medRxiv article using its DOI.

    Args:
        doi: DOI of the article

    Returns:
        Dictionary containing article metadata
    """
    try:
        metadata = await asyncio.to_thread(doi_get_medrxiv_metadata, doi)
        return metadata if metadata else {"error": f"No metadata found for DOI: {doi}"}
    except Exception as e:
        return {"error": f"An error occurred while fetching metadata: {str(e)}"}

'''
@app.tool()
async def search_medrxiv(prompt: str="Covid-19", num_results: int=3) -> Dict[str, Any]:
    print(f"In MCP Server search_medrxiv: {prompt}")
    results = await search_medrxiv_key_words(prompt)
    return results
'''

@app.prompt(description="Search medrxiv with prompts")
def get_initial_prompts() -> list[base.Message]:
    return [
        base.UserMessage("You are a helpful assistant that can help with medical research and clinical related questions."),
    ]


def create_starlette_app(mcp_server: Server, *, debug: bool = False) -> Starlette:
    """Create a Starlette application that can server the provied mcp server with SSE."""
    sse = SseServerTransport("/messages/")

    async def handle_sse(request: Request) -> None:
        async with sse.connect_sse(
                request.scope,
                request.receive,
                request._send,  # noqa: SLF001
        ) as (read_stream, write_stream):
            await mcp_server.run(
                read_stream,
                write_stream,
                mcp_server.create_initialization_options(),
            )

    return Starlette(
        debug=debug,
        routes=[
            Route("/sse", endpoint=handle_sse),
            Mount("/messages/", app=sse.handle_post_message),
        ],
    )


try:
    app.run(transport="streamable-http", host="0.0.0.0", port=8090)
finally:
    logging.info('Shutting down server medrxiv_server...')
'''
if __name__ == "__main__":
    mcp_server = app._mcp_server  # noqa: WPS437

    import argparse
    
    parser = argparse.ArgumentParser(description='Run MCP SSE-based server')
    parser.add_argument('--host', default='0.0.0.0', help='Host to bind to')
    parser.add_argument('--port', type=int, default=8083, help='Port to listen on')
    args = parser.parse_args()

    # Bind SSE request handling to MCP server
    starlette_app = create_starlette_app(mcp_server, debug=True)

    uvicorn.run(starlette_app, host=args.host, port=args.port)
'''


from openai import OpenAI
from mcp_master.config import ConfigError


# --------------------------------------------------------------------------------------------
# Helper Functions ---------------------------------------------------------------------------
# --------------------------------------------------------------------------------------------


def openai_url_invoke(model_id: str, user_query: str, prompt: str,
                      system_prompt: str = 'You are a seasoned expert.', service_url: str = '', max_token: int = 10240,
                      temperature: float = 0.05, top_p: float = 0.9):
    if model_id is None or len(model_id) == 0:
        raise ConfigError('Ensure your judge_model_id is properly configured via set_config().')
    if service_url is None or len(service_url) == 0:
        raise ConfigError('Ensure your judge_model_service_url is properly configured via set_config().')

    print(f'Invoking model "{model_id}" from "{service_url}"...')

    client = OpenAI(
        base_url=service_url,
    )

    chat_response = client.chat.completions.create(
        model=model_id,
        messages=[
            {"role": "system", "content": f'{system_prompt} Question: {user_query}'},
            {"role": "user", "content": prompt}
        ],
        stream=False,
        temperature=temperature,
        max_tokens=max_token,
        top_p=top_p
    )

    footer = f'Completion_tokens: {chat_response.usage.completion_tokens}, Prompt_tokens: {chat_response.usage.prompt_tokens},  Total_tokens:{chat_response.usage.total_tokens}'
    return chat_response.choices[0].message.content, footer
